--#include "../Prelude.agda"
--#include "../Relation.agda"


Setoid :: Type
Setoid =  sig 
          Elem :: Set
          Equal :: Rel Elem
          equivRelEq :: EquivRel Equal

data El (X::Setoid) ::  Set
     = ElOf (el :: X.Elem)

toEl (X::Setoid) :: El X -> X.Elem
toEl h = case h of 
             (ElOf el) -> el 


data Eq (|X::Setoid)(x1,x2::El X) :: Set 
    = isEq (h :: X.Equal (toEl x1 ) (toEl x2 ))


eqS (X::Setoid) :: (x1,x2::El X) -> Eq x1 x2 -> X.Equal (toEl x1) (toEl x2)
eqS h = \(x2::El X) -> \(h'::Eq h x2) -> case h' of 
                                             (isEq h0) -> h0


  
NotEq (|X::Setoid) :: Rel (El X)
NotEq x1 x2 = Not (Eq |X x1 x2)

Respectable (|X::Setoid)(P::Pred (El X)) :: Set
Respectable = (x1,x2::El X) |-> Eq x1 x2 -> P x1 -> P x2

RspEq (|X,|Y::Setoid):: (f::El X -> El Y) -> Set
RspEq f = (x1,x2::El X)  |-> Eq x1 x2 -> Eq (f x1) (f x2)


RspEq2 (|X,|Y,|Z::Setoid) :: (f::El X -> El Y -> El Z) -> Set
RspEq2 f = (x1,x2:: El X) |-> (y1,y2:: El Y) -> 
         Eq x1 x2 -> 
         Eq y1 y2  -> 
         Eq (f x1 y1) (f x2 y2) 


eqRefl(X::Setoid) :: Reflexive Eq
eqRefl = \(x::El X) -> isEq (X.equivRelEq.isRefl (toEl x))


eqSymm(X::Setoid) :: Symmetrical Eq
eqSymm = \(x1,x2::El X) |-> \(h::Eq x1 x2) -> 
         isEq (X.equivRelEq.isSymm (eqS x1 x2 h))


eqTrans(X::Setoid) :: Transitive Eq
eqTrans = \(x1,x2,x3::El X) |-> \(h::Eq x1 x2) -> \(h'::Eq x2 x3) -> 
         isEq (X.equivRelEq.isTrans (eqS x1 x2 h) (eqS x2 x3 h'))


TimesS :: (A,B::Setoid) -> Setoid
TimesS A B = struct 
               Elem ::  Set
               Elem  = El A  El B 
               Equal ::  Rel Elem
               Equal  = \(h,h'::El A  El B) -> 
                        Eq (fst h)(fst h')  Eq (snd h) (snd h')
               equivRelEq ::  EquivRel Equal
               equivRelEq  = struct 
                   isRefl ::  Reflexive Equal
                   isRefl  = \(x::El A  El B) -> (eqRefl (fst x), eqRefl (snd x))
                   isSymm ::  Symmetrical Equal
                   isSymm  = \(x1,x2::El A  El B) |-> \(h::Eq (fst x1) (fst x2)  Eq (snd x1) (snd x2)) -> (eqSymm (fst h), eqSymm (snd h))
                   isTrans ::  Transitive Equal
                   isTrans  = \(x1,x2,x3::El A  El B) |-> 
                              \(h::Eq (fst x1) (fst x2)  Eq (snd x1) (snd x2)) -> 
                              \(h'::Eq (fst x2) (fst x3)  Eq (snd x2) (snd x3)) -> 
                              (eqTrans (fst h) (fst h'), eqTrans (snd h) (snd h'))


FunS ::  (X,Y::Setoid) -> Set
FunS X Y = sig 
           app :: El X -> El Y
           rsp :: RspEq |X |Y app

appFun (X,Y::Setoid) :: (f::FunS X Y) -> El X -> El Y
appFun f = f.app

rspFun (X,Y::Setoid) :: (f::FunS X Y) -> RspEq (appFun f)
rspFun f = f.rsp
 
