--#include "Prelude.agda"
--#include "List.agda"
--#include "AlgLaw.agda"
--#include "Equalities/DecidableEquality.agda"
--#include "PropRelation.agda"




EqList(X::Set) :: (eqX :: HasDecEq X) -> Rel (List X)
EqList eqX = rel2Rel (eqList eqX.eq)

decEqList(X::Set) :: (eqX :: HasDecEq X) -> EqualityRel (EqList eqX)
decEqList = \(eqX::HasDecEq X) -> struct 
    isRefl ::Reflexive (EqList eqX)
           = \(x::List X) -> 
             case x of 
                    (Nil) -> isTrue
                    (x' : xs) -> andLem1 (eqX.isEq.isRefl x') (isRefl xs)
    isSubst ::Substitutive (EqList eqX)
       = \(P::List X -> Set) -> 
         \(x1,x2::List X) |-> 
         \(h:: EqList eqX x1 x2) -> 
         \(h'::P x1) -> case x1 of 
             (Nil) -> case x2 of 
                              (Nil) -> h'
                              (x : xs) -> absurdElim h isFalse (P (x : xs))
             (x : xs) -> case x2 of 
                         (Nil) -> absurdElim h isFalse (P [])
                         (x' : xs') -> let h1 :: IsTrue (eqX.eq x x' )
                                           h1 = andLem21  h
                                           h2 :: EqList eqX xs xs'
                                           h2 = andLem22 |(eqX.eq x x') h 
                                       in  eqX.isEq.isSubst 
                                             (\(x0::X) -> P (x0 : xs'))
                                             h1
                                             (isSubst (\xs -> P (x:xs)) h2 h')

hasDecEqList(X::Set) :: (eqX :: HasDecEq X) -> HasDecEq (List X)
hasDecEqList eqX = struct 
                     eq ::  List X -> List X -> Bool
                     eq  = eqList eqX.eq
                     isEq ::  EqualityRel (rel2Rel eq)
                     isEq  = decEqList eqX



reflEqList (X::Set) :: (eqX :: HasDecEq X) -> Reflexive (EqList eqX)
reflEqList eqX = reflDecEq (hasDecEqList eqX)

symmEqList (X::Set) :: (eqX :: HasDecEq X) -> Symmetrical (EqList eqX)
symmEqList eqX = symmDecEq (hasDecEqList eqX)

transEqList(X::Set) :: (eqX :: HasDecEq X) -> Transitive (EqList eqX)
transEqList eqX = transDecEq (hasDecEqList eqX)

substEqList(X::Set) :: (eqX :: HasDecEq X) -> Substitutive(EqList eqX)
substEqList eqX = substDecEq (hasDecEqList eqX)


distrConsEqListH(X::Set) 
      (eqX :: HasDecEq X)
      (x1,x2::X)
      (xs::List X) ::
      EqList eqX (x1:xs) (x2:xs) ->
      decEq eqX x1 x2 
distrConsEqListH eql = andLem21 eql


distrConsEqListT(X::Set) 
      (eqX :: HasDecEq X)
      (x::X)
      (xs1,xs2::List X) ::
      EqList eqX (x:xs1) (x:xs2) ->
      decEq (hasDecEqList eqX) xs1 xs2
distrConsEqListT eql = andLem22 |(eqX.eq x x) eql


eqList2Id (X::Set)
          (eqX :: HasDecEq X)
       :: (xs,ys::List X) ->
          EqList eqX xs ys ->
          xs == ys
eqList2Id xs ys eql = equalityRelId (EqList eqX) (decEqList eqX) xs ys eql


id2EqList (X::Set)
          (eqX :: HasDecEq X)
          (xs,ys::List X) 
       :: xs == ys -> 
          EqList eqX xs ys
id2EqList idl = idReflRel (EqList eqX) xs ys (reflEqList eqX) idl

