--#include "Prelude.agda"
--#include "Bool.agda"
--#include "Nat.agda" 
--#include "Maybe.agda"

---------------------------------------------------------------------------
-- Lists. -----------------------------------------------------------------
---------------------------------------------------------------------------

elimList (X::Set) :: 
         (C::List X -> Set) ->
         (c_nil::C []) ->
         (c_con::(x::X) -> (xs::List X) -> C xs -> C (x:xs)) ->
         (xs::List X) ->
         C xs
elimList C c_nil c_con xs = case xs of 
                                  (Nil) -> c_nil
                                  (x : xs') ->
                                    c_con x xs' (elimList  C c_nil c_con xs')

foldl (X,Y::Set) ::  (X -> Y -> X) -> X -> List Y -> X
foldl f x ys = 
            case ys of 
                (Nil) -> x
                (y:ys') -> foldl f (f x y) ys'

foldr (X,Y::Set) ::  (X -> Y -> Y) -> Y -> List X -> Y
foldr f y xs =
            case xs of 
                (Nil) -> y
                (x : xs') -> f x (foldr f y xs')

map (X,Y::Set) ::  (X -> Y) -> List X -> List Y
map f xs = case xs of 
                 (Nil) -> []
                 (x : xs') -> f x : map f xs'  


null (X::Set) :: List X -> Bool
null xs = case xs of 
              (Nil) -> True
              (x : xs') -> False

eqList(X::Set) :: (eqX :: X -> X -> Bool) -> List X -> List X -> Bool
eqList eqX xs ys = 
    case xs of 
        (Nil) -> null ys
        (x : xs') -> case ys of 
                         (Nil) -> False
                         (y : ys') -> eqX x y && eqList eqX xs' ys'



(++)(a::Set) :: List a -> List a -> List a
xs ++ ys = foldr (:) ys xs

filter(a::Set) :: (a -> Bool) -> List a -> List a
filter p xs = case xs of
              (Nil) -> []
              (x:xs) -> if p x then x : filter p xs
                          else filter p xs


concat(a::Set) :: List (List a) -> List a
concat xss = foldr (++) [] xss
               

concatMap(a,b::Set) :: (a -> List b) -> List a -> List b
concatMap f = concat  map f


idata NotNil (|A::Set) :: List A -> Set where
        notNil(x::A)(xs::List A) :: NotNil (x:xs)
  

notNilElim (A::Set) :: 
      (C :: (l::List A) -> NotNil l -> Set) -> 
      (c :: (x::A) -> (xs :: List A) -> C (x:xs) (notNil x xs)) -> 
      (l :: List A) -> (h :: NotNil l) -> C l h
  
notNilElim C c l h = case h of 
                       (notNil x xs) -> c x xs

NotNull(a::Set) :: List a -> Set
NotNull xs = case xs of
               (Nil) -> Empty
               (x:xs') -> Unit


notNotNil (a::Set) :: Not (NotNil |a []) 
notNotNil = \(h::NotNil []) -> 
              notNilElim (\l h -> NotNull l) (\x -> \xs -> unit) [] h


-- head and tail extract the first element and remaining elements,
-- respectively, of a list, which must be non-empty.  last and init
-- are the dual functions working from the end of a finite list,
-- rather than the beginning.


head(a::Set)     :: (l::List a) -> NotNil l -> a
head xs p = case xs of 
              (Nil) -> absurdElim p notNotNil a
              (x : xs') -> x

  
tail(a::Set)     :: (l::List a) -> NotNil l -> List a
tail xs p  = case xs of 
               (Nil) -> absurdElim p notNotNil (List a)
               (x : xs') -> xs'

  

last(a::Set)    :: (l::List a) -> NotNil l -> a
last xs p = case xs of 
              (Nil) -> absurdElim p notNotNil a
              (x : xs') -> 
                  case xs' of 
                      (Nil) -> x
                      (x' : xs0) -> last xs' (notNil x' xs0) 

  
init(a::Set)    :: (xs::List a) -> NotNil xs -> List a
init xs p = case xs of 
              (Nil) -> absurdElim p notNotNil (List a)
              (x : xs') -> 
                  case xs' of 
                      (Nil) -> [] 
                      (x' : xs0) -> x : init xs' (notNil x' xs0) 
  


length(a::Set)  :: List a -> Nat
length l = case l of 
             (Nil) -> zero
             (x : xs) -> succ (length xs)


foldl1(a::Set)    :: (a -> a -> a) -> (l::List a) -> NotNil l -> a
foldl1 f xs p = case xs of 
                  (Nil) -> absurdElim p notNotNil a 
                  (x : xs') -> foldl f x xs' 

scanl(a,b::Set)   :: (a -> b -> a) -> a -> List b -> List a
scanl f q xs     = case xs of 
                     (Nil) -> []
                     (x : xs') -> q : scanl f (f q x) xs'


scanl1 (a::Set)  :: (a -> a -> a) -> (xs :: List a) ->  NotNil xs -> List a
scanl1 f xs p = case xs of 
                  (Nil) -> absurdElim p notNotNil (List a)
                  (x : xs') -> scanl f x xs'



foldr1(a::Set)    :: (a -> a -> a) -> (xs :: List a) -> NotNil xs -> a
foldr1 f xs p = 
        case xs of 
        (Nil) -> absurdElim p notNotNil a
        (x : xs') -> 
            case xs' of 
                 (Nil) -> x
                 (x' : xs0) -> f x (foldr1 f xs' (notNil x' xs0))
  

scanr(a,b::Set)   :: (a -> b -> b) -> b -> List a -> Sigma (\l -> NotNil l)
scanr f q0 xs = 
       case xs of 
           (Nil) -> dep_pair [q0] (notNil q0 [])
           (x : xs') -> 
               let qs ::Sigma (\(l::List b) -> NotNil l)
                   qs = scanr f q0 xs'
                   q :: b 
                   q = head (dep_fst qs) (dep_snd qs)
                   p :: NotNil (f x q : dep_fst qs)
                   p = notNil (f x q) (dep_fst qs)
               in  dep_pair (f x q : dep_fst qs) p

 
scanr1(a::Set)   :: (a -> a -> a) -> (xs::List a) -> NotNil xs -> Sigma (\(l::List a) -> NotNil l)
scanr1 f xs p = 
      case xs of 
          (Nil) -> absurdElim p notNotNil (Sigma (\(l::List a) -> NotNil l))
          (x : xs') -> 
              case xs' of 
                  (Nil) -> dep_pair [x] (notNil x [])
                  (x' : xs0) -> 
                      let qs ::Sigma (\(l::List a) -> NotNil l)
                          qs = scanr1 f xs' (notNil x' xs0)
                          l :: List a
                          l = f x (head (dep_fst qs) (dep_snd qs)) : dep_fst qs 
                      in  dep_pair l (notNil (f x (head (dep_fst qs) (dep_snd qs))) (dep_fst qs))

take(a::Set)           :: (n::Nat) -> (xs::List a) ->  List a
take n xs = case xs of 
                  (Nil) -> [] 
                  (x : xs') -> case n of 
                                   (zero) -> [] 
                                   (succ m) -> take m xs'  

drop(a::Set)           :: (n::Nat) -> (xs::List a) ->  List a
drop n xs = case n of 
                  (zero) -> xs
                  (succ m) -> case xs of 
                                  (Nil) -> [] 
                                  (x : xs') -> drop m xs' 


splitAt(a::Set)          :: Nat -> List a -> List a  List a
splitAt n xs = case n of 
                     (zero) -> ([],xs) 
                     (succ m) -> case xs of 
                                     (Nil) -> ([],[]) 
                                     (x : xs') -> let spl ::List a  List a
                                                        = splitAt m xs' 
                                                  in  (x:fst spl,snd spl) 

takeWhile(a::Set)        :: (a -> Bool) -> List a -> List a
takeWhile p xs = case xs of 
                       (Nil) -> []
                       (x : xs') -> if p x then x:takeWhile p xs' else [] 


dropWhile(a::Set)        :: (a -> Bool) -> List a -> List a
dropWhile p xs = case xs of 
                       (Nil) -> [] 
                       (x : xs') -> if p x then dropWhile p xs' else xs 

span(a::Set )    :: (a -> Bool) -> List a -> List a  List a
span p xs = case xs of
                  (Nil) ->  ([],[])
                  (x:xs') -> let spn :: List a  List a 
                                 spn = span p xs'
                             in if p x then (x:fst spn, snd spn) else ([],xs)

break(a::Set )    :: (a -> Bool) -> List a -> List a  List a
break p                 =  span (not  p)
  
and         :: List Bool  -> Bool
and              =  foldr (&&) True
  
or         :: List Bool -> Bool
or          =  foldr (||) False


any(a::Set) :: (a -> Bool) -> List a -> Bool
any p            =  or  map p

all(a::Set) :: (a -> Bool) -> List a -> Bool
all p            =  and  map p

-- elem is the list membership predicate, usually written in infix form,
-- e.g., x `elem` xs.  notElem is the negation.

elem (a::Set)    :: ((===) :: a -> a -> Bool) -> a -> List a -> Bool
elem (===) x           =  any ((===) x)

notElem(a::Set)    :: ((===) :: a -> a -> Bool) -> a -> List a -> Bool
notElem (===) x        =  all (\(y::a) -> not (x===y))


  
lookup(a,b::Set)   :: (a -> a -> Bool) -> a -> List (ab) -> Maybe b
lookup p x xs  = 
      case xs of 
          (Nil) -> Nothing
          (x' : xs') -> 
              case x' of 
                  (a',b') -> if p x a' then Just b' else lookup p x xs' 

idata IndexOfList (A::Set) :: List A -> Nat -> Set where
      index0(x::A)(xs::List A) :: IndexOfList (x:xs) zero
      indexS(x::A)(xs::List A)(n::Nat)(h::IndexOfList xs n) :: IndexOfList (x:xs) (succ n)

type Index(|A::Set)(xs::List A) = Sigma (\(n::Nat) -> IndexOfList xs n)

(!!)(a::Set) :: (xs::List a) -> Index xs -> a
xs !! p = case p of 
                (dep_pair x y) -> case y of 
                                      (index0 x' xs') -> x'
                                      (indexS x' xs' n h) -> xs' !! (dep_pair n h) 

zipWith(a,b,c::Set) :: (a->b->c) -> List a -> List b -> List c 
zipWith z xs ys = case xs of 
                        (Nil) -> []
                        (x : xs') -> case ys of 
                                         (Nil) -> []
                                         (y : ys') -> z x y : zipWith z xs' ys'  
zip(a,b::Set)    :: List a -> List b -> List (ab)
zip              =  zipWith (,)


unzip(a,b::Set)  :: List (ab) -> List a  List b
unzip  =  foldr (\ab abs -> (fst ab : fst abs,snd ab :snd abs)) ([],[])
