-- Inbuilt prelude definitions  -------------------------------------------
{- 
data ()(X,Y::Set) = (,)(a::A)(b::B)
data List (X::Set) = Nil | (:)(x::X)(xs::List X)
data Bool= False | True
if b then e else e' 
if_then_else (X::Set) :: Bool -> X -> X -> X ...
type Char = ...
type Integer = ...
type String = List Char
-}

-----------------------------------------------------------------------
-- Some standard function ---------------------------------------------
-----------------------------------------------------------------------

id (X::Set) ::  X -> X
id  = \x -> x

() (X,Y,Z::Set) ::  (Y -> Z) -> (X -> Y) -> X -> Z
()  = \f -> \g -> \x -> f (g x)

seq (X,Y,Z::Set) ::  (X -> Y) -> (Y -> Z) -> X -> Z
seq f g = g  f

const (X,Y::Set) :: X -> Y -> X
const x y = x

proj (X::Set) ::  (Y::X -> Set) -> (x::X) -> (f::(x::X) -> Y x) -> Y x
proj Y x f = f x

flip (X,Y,Z::Set) :: (X -> Y -> Z) -> Y -> X -> Z
flip f y x = f x y

($)(X,Y::Set) :: (X -> Y) -> X -> Y
f $  x    =  f x



Pred :: Set -> Type
Pred X = X -> Set

Rel ::  Set -> Type
Rel X = X -> X -> Set

---------------------------------------------------------------------------
--- The empty set  --------------------------------------------------------
---------------------------------------------------------------------------
-- The empty set corresponds to absurdity ---------------------------------

data Empty :: Set = 

elimEmpty ::  (C::Empty -> Set) -> (z::Empty) -> C z
elimEmpty C z = case z of { }

whenEmpty ::  (X::Set) -> Empty -> X
whenEmpty X z = case z of { }

Not :: Set -> Set 
Not X = X -> Empty
  
absurdElim (A::Set) :: A -> Not A -> (X::Set) -> X
absurdElim h h' X = whenEmpty X (h' h)

---------------------------------------------------------------------------
--- The unit set  ---------------------------------------------------------
---------------------------------------------------------------------------
-- The unit set corresponds to truth --------------------------------------

data Unit :: Set  = unit

elimUnit ::  (C::Unit -> Set) -> (c_unit::C unit) -> (u::Unit) -> C u
elimUnit C c_unit u = case u of 
                            (unit) -> c_unit

---------------------------------------------------------------------------
--- Tuple -----------------------------------------------------------------
---------------------------------------------------------------------------
-- Corresponds to and -----------------------------------------------------

elimPair (X,Y::Set) :: (C :: (X  Y) -> Set) -> 
                       (c_p :: (x::X) -> (y::Y) -> C (x,y)) ->
                       (p :: X  Y) ->
                       C p
elimPair C c_p p = 
    case p of 
        (x,y) -> c_p x y

whenPair (X,Y,Z::Set) :: (X -> Y -> Z) -> X  Y -> Z
whenPair c_p p = 
      case p of 
          (x,y) -> c_p x y

fst(X,Y::Set) :: XY -> X
fst p = case p of (x,y) -> x

snd(X,Y::Set) :: XY -> Y
snd p = case p of (x,y) -> y

curry (X,Y,Z::Set) ::  (X  Y -> Z) -> (X -> Y -> Z)
curry f = \(x::X) -> \(y::Y) -> f (x,y)
  
uncurry (X,Y,Z::Set) :: (X -> Y -> Z) -> (X  Y -> Z)
uncurry = whenPair
  
curryS (X,Y::Set) :: (Z::X  Y -> Set) -> 
    ((xy:: X  Y) -> Z xy) -> (x::X) -> (y::Y) -> Z (x,y)
curryS Z f x y = f (x,y)

uncurryS (X::Set)(Y::Set) :: (Z::X  Y -> Set) -> 
    ((x::X) -> (y::Y) -> Z (x,y)) -> (xy::X  Y) -> Z xy
uncurryS = elimPair

mapPair (X1,X2,Y1,Y2::Set) :: (X1 -> Y1) -> (X2 -> Y2) -> X1X2 -> Y1Y2
mapPair f g p = 
      case p of
          (x1,x2) -> (f x1,g x2) 

---------------------------------------------------------------------------
-- The (binary) disjoint union. -------------------------------------------
---------------------------------------------------------------------------
-- Corresponds to or ------------------------------------------------------

data Plus (X,Y::Set)  = Inl (x::X) | Inr (y::Y)

elimPlus (X,Y::Set) :: 
    (C::Plus X Y -> Set) ->
    (c_lft::(x::X) -> C (Inl x)) ->
    (c_rgt::(y::Y) -> C (Inr y)) ->
    (xy::Plus X Y) ->
    C xy
elimPlus C c_lft c_rgt xy = case xy of 
                                  (Inl x) -> c_lft x
                                  (Inr y) -> c_rgt y

whenPlus (X,Y,Z::Set) ::  (f::X -> Z) -> (g::Y -> Z) -> (Plus X Y -> Z)
whenPlus f g = \xy ->
             case xy of 
                 (Inl x) -> f x
                 (Inr y) -> g y

mapPlus (X1,X2,Y1,Y2::Set) :: 
    (f::X1 -> X2) -> (g::Y1 -> Y2) -> (Plus X1 Y1 -> Plus X2 Y2)
mapPlus f g = whenPlus (\x1 -> Inl (f x1)) (\y1 -> Inr (g y1))



---------------------------------------------------------------------------
-- Dependent pairs. -------------------------------------------------------
---------------------------------------------------------------------------
-- Corresponds to existential ---------------------------------------------



data Sigma (|X::Set) (Y::X -> Set) = 
       dep_pair (x::X)(y::Y x)
               
dep_fst (X::Set)(Y::(X -> Set)) ::  (xy::Sigma Y) -> X
dep_fst xy = case xy of 
                   (dep_pair x y) -> x

dep_snd (X::Set)(Y::(X -> Set)) ::  (xy::Sigma Y) -> Y (dep_fst xy)
dep_snd xy = case xy of 
                   (dep_pair x y) -> y

dep_curry (X::Set)(Y::(X -> Set))(Z::Set) :: 
    (f::Sigma (\x -> Y x) -> Z) -> (x::X) |-> (Y x -> Z)
dep_curry f = \x |-> \y -> f (dep_pair x y)

dep_uncurry (X::Set)(Y::(X -> Set))(Z::Set) :: 
    ((x::X) -> (Y x -> Z)) -> Sigma (\x -> Y x) -> Z
dep_uncurry  f xy = f (dep_fst xy) (dep_snd xy)

dep_curryS (X::Set)(Y::(X -> Set)) :: 
    (Z::Sigma (\x -> Y x) -> Set) ->
    (f::(xy::Sigma (\x -> Y x)) -> Z xy) ->
    (x::X) ->
    (y::Y x) ->
    Z (dep_pair x y)
dep_curryS Z f = \(x::X) -> \(y::Y x) -> f (dep_pair x y)

dep_uncurryS (X::Set)(Y::(X -> Set)) :: 
    (Z::Sigma (\x -> Y x) -> Set) ->
    (f::(x::X) -> (y::Y x) -> Z (dep_pair x y)) ->
    (xy::Sigma (\x -> Y x)) ->
    Z xy
dep_uncurryS Z f xy = case xy of 
                            (dep_pair x y) -> f x y 

mapSigma (X::Set)(Y1,Y2::X -> Set) :: 
    (f::(x::X) -> (Y1 x -> Y2 x)) -> Sigma (\x -> Y1 x) -> Sigma (\x -> Y2 x)
mapSigma f p = case p of 
                     (dep_pair x y) -> dep_pair x (f x y) 



---------------------------------------------------------------------------
-- Identity proof sets. ---------------------------------------------------
---------------------------------------------------------------------------

idata (==) (X::Set)  :: X  -> X -> Set where
       refId (x::X) :: (==) x  x

elimId (X::Set) :: 
    (C::(x1,x2::X) |-> x1 == x2 -> Set) ->
    (refC::(x::X) -> C (refId x)) ->
    (x1,x2::X) |->
    (u::x1 == x2) ->
    C u
elimId C refC |x1 |x2 u = case u of 
                                (refId x) -> refC x

whenId (X::Set) :: 
       (C:: X -> X -> Set) ->
       (c::(x::X) -> C x x) ->
       (x1,x2::X) |->
       x1 == x2 -> 
       C x1 x2
whenId C c = elimId (\x1 x2 |-> \(u::x1 == x2) -> C x1 x2) c
  
mapId (X,Y::Set) :: 
             (f::X -> Y) -> (x1,x2::X) |-> x1 == x2 -> (f x1) == (f x2)
  
mapId f = whenId (\x1 x2 -> (f x1) == (f x2)) (\(x::X) -> refId (f x))



---------------------------------------------------------------------------
-- Function sets ----------------------------------------------------------
--------------------------------------------------------------------------

data Fun (X,Y::Set) = fun (f :: X -> Y)

apply (X,Y::Set) :: Fun X Y -> X -> Y
apply f x = case f of 
                  (fun f') -> f' x

mapFun (X1,X2,Y1,Y2::Set) :: 
    (X2 -> X1) -> (Y1 -> Y2) -> Fun X1 Y1 -> Fun X2  Y2
mapFun  f g h = case h of 
                      (fun h') -> fun (\(x::X2) -> g (h' (f x)))



----------------------------------------------------------------------------
-- Product sets.
----------------------------------------------------------------------------

data Prod (X::Set)(Y::X -> Set) = prod (f :: (x::X) -> Y x)
 
app(X::Set)(Y::X -> Set) :: Prod X Y -> (x::X) -> Y x
app f x = case f of 
                (prod f') -> f' x
 
mapProd (X::Set)(Y1,Y2::X -> Set) :: 
    (f::(x::X) -> Y1 x -> Y2 x) -> Prod X Y1 -> Prod X Y2
mapProd f p = case p of 
                    (prod f') -> prod (\(x::X) -> f x (f' x))

  




--------------------------------------------------------------------------
---- Some standard properties --------------------------------------------
--------------------------------------------------------------------------


Decidable :: (X::Set) -> Set
Decidable X = Plus X (Not X)
  
DecidablePred (|X::Set) :: (P::Pred X) -> Set
DecidablePred P = (x::X) -> Decidable (P x)


Collapsed :: (X::Set) -> Type
Collapsed X = (P::Pred X) -> (x1,x2::X) |-> P x1 -> P x2

