--#include "Prelude.agda"

Reflexive (X::Set) ::  (R::Rel X) -> Set
Reflexive R = (x::X) -> x `R` x

Symmetrical (X::Set) ::  (R::Rel X) -> Set
Symmetrical R = (x1,x2::X) |-> (x1 `R` x2 -> x2 `R` x1)

Transitive (X::Set) ::  (R::Rel X) -> Set
Transitive R = (x1,x2,x3::X) |-> (x1 `R` x2 -> x2 `R` x3 -> x1 `R` x3)
  
Compositional (X::Set) ::  (R::Rel X) -> Set
Compositional R = (x1,x2,x3::X) |->
                    x2 `R` x3 -> x1 `R` x2 -> x1 `R` x3

Substitutive (X::Set) ::  (R::Rel X) -> Type
Substitutive R = (P::X -> Set) ->
                 (x1,x2::X) |->
                 x1 `R` x2 -> P x1 -> P x2

DecidableRel (|X::Set) :: (R::Rel X) -> Set
DecidableRel R = (x1,x2::X)  -> Decidable (R x1 x2)


FlipRel (X::Set) ::  (R::Rel X) -> X -> X -> Set
FlipRel R x1 x2 = x2 `R` x1


()(X::Set) :: (R::Rel X) -> (S::Rel X) -> Rel X
R  S = \ a b ->
    Sigma (\(a'::X) -> (a `R` a')  (S a' b))

subRel(X::Set):: (R::Rel X) -> (S::Rel X) -> Set
subRel R S = (a,a':: X) -> a `R` a' -> S a a'


unionRel (X::Set) :: (R::Rel X) -> (S::Rel X) -> Rel X
unionRel R S = \(a,b::X) -> (a `R` b) `Plus` (S a b)


interRel(X::Set) :: (R::Rel X) -> (S::Rel X) -> Rel X
interRel R S = \a b -> (a `R` b)  (a `S `b)



equalRel(X::Set) :: (R::Rel X) -> (S::Rel X) -> Set
equalRel R S = (subRel R S)  (subRel S R)


EquivRel(X::Set) :: (R::Rel X) -> Set
EquivRel R = sig 
     isRefl :: Reflexive R
     isSymm :: Symmetrical R
     isTrans :: Transitive R


EqualityRel(X::Set) :: (R::Rel X) -> Type
EqualityRel R = sig 
     isRefl :: Reflexive R
     isSubst:: Substitutive R


equalityRelEquiv(X::Set) :: (R::Rel X) -> EqualityRel R -> EquivRel R
equalityRelEquiv R eqR = struct 
    isRefl ::  Reflexive R
    isRefl  = eqR.isRefl
    isSymm ::  Symmetrical R
    isSymm  = \(x1,x2::X) |-> \(h::x1 `R` x2) ->
              eqR.isSubst (\x2'-> x2' `R` x1) h (isRefl x1)
    isTrans ::  Transitive R
    isTrans  = \(x1,x2,x3::X) |-> 
               \(h::x1 `R` x2) -> 
               \(h'::x2 `R` x3) -> 
               eqR.isSubst (R x1) h' h




